def format_error_message(mapping, op):
    sep = "\n    " if len(mapping) == 1 else "\n -- "
    if op == "attach":
        message = "Cannot attach units:"
        message = sep.join(
            [message]
            + [
                f"cannot attach units to variable {key!r}: {unit} (reason: {str(e)})"
                for key, (unit, e) in mapping.items()
            ]
        )
    elif op == "parse":
        message = "Cannot parse units:"
        message = sep.join(
            [message]
            + [
                f"invalid units for variable {key!r}: {unit} ({type}) (reason: {str(e)})"
                for key, (unit, type, e) in mapping.items()
            ]
        )
    elif op == "convert":
        message = "Cannot convert variables:"
        message = sep.join(
            [message]
            + [
                f"incompatible units for variable {key!r}: {error}"
                for key, error in mapping.items()
            ]
        )
    elif op == "convert_indexers":
        message = "Cannot convert indexers:"
        message = sep.join(
            [message]
            + [
                f"incompatible units for indexer for {key!r}: {error}"
                for key, error in mapping.items()
            ]
        )
    else:
        raise ValueError("invalid op")

    return message



import pint
import pytest

# only need to register _repr_inline_
import pint_xarray  # noqa: F401

unit_registry = pint.UnitRegistry(force_ndarray_like=True)


@pytest.mark.parametrize(
    ("length", "expected"),
    (
        (40, "[N] 7.1 5.4 9.8 21.4 15.3"),
        (20, "[N] 7.1 5.4 ... 15.3"),
        (10, "[N] 7.1..."),
        (7, "[N] ..."),
        (3, "[N] ..."),
    ),
)
def test_inline_repr(length, expected):
    quantity = unit_registry.Quantity([7.1, 5.4, 9.8, 21.4, 15.3], "N")

    assert quantity._repr_inline_(length) == expected



from itertools import zip_longest

import numpy as np
from xarray.core.options import OPTIONS


# vendored from xarray.core.formatting
def maybe_truncate(obj, maxlen=500):
    s = str(obj)
    if len(s) > maxlen:
        s = s[: (maxlen - 3)] + "..."
    return s


# vendored from xarray.core.formatting
def pretty_print(x, numchars: int):
    """Given an object `x`, call `str(x)` and format the returned string so
    that it is numchars long, padding with trailing spaces or truncating with
    ellipses as necessary
    """
    s = maybe_truncate(x, numchars)
    return s + " " * max(numchars - len(s), 0)


# vendored from xarray.core.formatting
def _get_indexer_at_least_n_items(shape, n_desired, from_end):
    assert 0 < n_desired <= np.prod(shape)
    cum_items = np.cumprod(shape[::-1])
    n_steps = np.argmax(cum_items >= n_desired)
    stop = int(np.ceil(float(n_desired) / np.r_[1, cum_items][n_steps]))
    indexer = (
        ((-1 if from_end else 0),) * (len(shape) - 1 - n_steps)
        + ((slice(-stop, None) if from_end else slice(stop)),)
        + (slice(None),) * n_steps
    )
    return indexer


# vendored from xarray.core.formatting
def first_n_items(array, n_desired):
    """Returns the first n_desired items of an array"""
    # Unfortunately, we can't just do array.flat[:n_desired] here because it
    # might not be a numpy.ndarray. Moreover, access to elements of the array
    # could be very expensive (e.g. if it's only available over DAP), so go out
    # of our way to get them in a single call to __getitem__ using only slices.
    if n_desired < 1:
        raise ValueError("must request at least one item")

    if array.size == 0:
        # work around for https://github.com/numpy/numpy/issues/5195
        return []

    if n_desired < array.size:
        indexer = _get_indexer_at_least_n_items(array.shape, n_desired, from_end=False)
        array = array[indexer]
    return np.asarray(array).flat[:n_desired]


# vendored from xarray.core.formatting
def last_n_items(array, n_desired):
    """Returns the last n_desired items of an array"""
    # Unfortunately, we can't just do array.flat[-n_desired:] here because it
    # might not be a numpy.ndarray. Moreover, access to elements of the array
    # could be very expensive (e.g. if it's only available over DAP), so go out
    # of our way to get them in a single call to __getitem__ using only slices.
    if (n_desired == 0) or (array.size == 0):
        return []

    if n_desired < array.size:
        indexer = _get_indexer_at_least_n_items(array.shape, n_desired, from_end=True)
        array = array[indexer]
    return np.asarray(array).flat[-n_desired:]


# based on xarray.core.formatting.format_item
def format_item(x, quote_strings=True):
    """Returns a succinct summary of an object as a string"""
    if isinstance(x, (str, bytes)):
        return repr(x) if quote_strings else x
    elif isinstance(x, (float, np.float_)):
        return f"{x:.4}"
    else:
        return str(x)


# based on xarray.core.formatting.format_item
def format_items(x):
    """Returns a succinct summaries of all items in a sequence as strings"""
    x = np.asarray(x)
    formatted = [format_item(xi) for xi in x]
    return formatted


def summarize_attr(key, value, col_width=None):
    """Summary for __repr__ - use ``X.attrs[key]`` for full value."""
    # Indent key and add ':', then right-pad if col_width is not None
    k_str = f"    {key}:"
    if col_width is not None:
        k_str = pretty_print(k_str, col_width)
    # Replace tabs and newlines, so we print on one line in known width
    v_str = str(value).replace("\t", "\\t").replace("\n", "\\n")
    # Finally, truncate to the desired display width
    return maybe_truncate(f"{k_str} {v_str}", OPTIONS["display_width"])


# adapted from xarray.core.formatting
def _diff_mapping_repr(a_mapping, b_mapping, title, summarizer, col_width=None):
    def extra_items_repr(extra_keys, mapping, ab_side):
        extra_repr = [summarizer(k, mapping[k], col_width) for k in extra_keys]
        if extra_repr:
            header = f"{title} only on the {ab_side} object:"
            return [header] + extra_repr
        else:
            return []

    a_keys = set(a_mapping)
    b_keys = set(b_mapping)

    summary = []

    diff_items = []

    for k in a_keys & b_keys:
        compatible = a_mapping[k] == b_mapping[k]
        if not compatible:
            temp = [
                summarizer(k, vars[k], col_width) for vars in (a_mapping, b_mapping)
            ]

            diff_items += [ab_side + s[1:] for ab_side, s in zip(("L", "R"), temp)]

    if diff_items:
        summary += [f"Differing {title.lower()}:"] + diff_items

    summary += extra_items_repr(a_keys - b_keys, a_mapping, "left")
    summary += extra_items_repr(b_keys - a_keys, b_mapping, "right")

    return "\n".join(summary)


# vendored from xarray.core.formatting
def format_array_flat(array, max_width: int):
    """Return a formatted string for as many items in the flattened version of
    array that will fit within max_width characters.
    """
    # every item will take up at least two characters, but we always want to
    # print at least first and last items
    max_possibly_relevant = min(
        max(array.size, 1), max(int(np.ceil(max_width / 2.0)), 2)
    )
    relevant_front_items = format_items(
        first_n_items(array, (max_possibly_relevant + 1) // 2)
    )
    relevant_back_items = format_items(last_n_items(array, max_possibly_relevant // 2))
    # interleave relevant front and back items:
    #     [a, b, c] and [y, z] -> [a, z, b, y, c]
    relevant_items = sum(
        zip_longest(relevant_front_items, reversed(relevant_back_items)), ()
    )[:max_possibly_relevant]

    cum_len = np.cumsum([len(s) + 1 for s in relevant_items]) - 1
    if (array.size > 2) and (
        (max_possibly_relevant < array.size) or (cum_len > max_width).any()
    ):
        padding = " ... "
        count = min(
            array.size, max(np.argmax(cum_len + len(padding) - 1 > max_width), 2)
        )
    else:
        count = array.size
        padding = "" if (count <= 1) else " "

    num_front = (count + 1) // 2
    num_back = count - num_front
    # note that num_back is 0 <--> array.size is 0 or 1
    #                         <--> relevant_back_items is []
    pprint_str = "".join(
        [
            " ".join(relevant_front_items[:num_front]),
            padding,
            " ".join(relevant_back_items[-num_back:]),
        ]
    )

    # As a final check, if it's still too long even with the limit in values,
    # replace the end with an ellipsis
    # NB: this will still returns a full 3-character ellipsis when max_width < 3
    if len(pprint_str) > max_width:
        pprint_str = pprint_str[: max(max_width - 3, 0)] + "..."

    return pprint_str


def inline_repr(quantity, max_width):
    magnitude = quantity.magnitude
    units = quantity.units

    units_repr = f"{units:~P}"
    if isinstance(magnitude, np.ndarray):
        data_repr = format_array_flat(magnitude, max_width - len(units_repr) - 3)
    else:
        data_repr = maybe_truncate(repr(magnitude), max_width - len(units_repr) - 3)

    return f"[{units_repr}] {data_repr}"



from importlib.metadata import version

import pint

from . import accessors, formatting, testing  # noqa: F401
from .accessors import default_registry as unit_registry
from .accessors import setup_registry

try:
    __version__ = version("pint-xarray")
except Exception:
    # Local copy or not installed with setuptools.
    # Disable minimum version checks on downstream libraries.
    __version__ = "999"


pint.Quantity._repr_inline_ = formatting.inline_repr


__all__ = [
    "testing",
    "unit_registry",
    "setup_registry",
]



import numpy as np
import pandas as pd
import pint
import pytest
import xarray as xr
from numpy.testing import assert_array_equal
from pint import Unit, UnitRegistry

from .. import accessors, conversion
from .utils import (
    assert_equal,
    assert_identical,
    assert_units_equal,
    requires_bottleneck,
    requires_dask_array,
    requires_scipy,
)

pytestmark = [
    pytest.mark.filterwarnings("error::pint.UnitStrippedWarning"),
]

# make sure scalars are converted to 0d arrays so quantities can
# always be treated like ndarrays
unit_registry = UnitRegistry(force_ndarray=True)
Quantity = unit_registry.Quantity

nan = np.nan


def assert_all_str_or_none(mapping):
    __tracebackhide__ = True

    compared = {
        key: isinstance(value, str) or value is None for key, value in mapping.items()
    }
    not_passing = {key: value for key, value in mapping.items() if not compared[key]}
    check = all(compared.values())

    assert check, f"Not all values are str or None: {not_passing}"


@pytest.fixture
def example_unitless_da():
    array = np.linspace(0, 10, 20)
    x = np.arange(20)
    u = np.linspace(0, 1, 20)
    da = xr.DataArray(
        data=array,
        dims="x",
        coords={"x": ("x", x), "u": ("x", u, {"units": "hour"})},
        attrs={"units": "m"},
    )
    return da


@pytest.fixture()
def example_quantity_da():
    array = np.linspace(0, 10, 20) * unit_registry.m
    x = np.arange(20)
    u = np.linspace(0, 1, 20) * unit_registry.hour
    return xr.DataArray(data=array, dims="x", coords={"x": ("x", x), "u": ("x", u)})


class TestQuantifyDataArray:
    def test_attach_units_from_str(self, example_unitless_da):
        orig = example_unitless_da
        result = orig.pint.quantify("s")
        assert_array_equal(result.data.magnitude, orig.data)
        # TODO better comparisons for when you can't access the unit_registry?
        assert str(result.data.units) == "second"

    def test_attach_units_given_registry(self, example_unitless_da):
        orig = example_unitless_da
        ureg = UnitRegistry(force_ndarray=True)
        result = orig.pint.quantify("m", unit_registry=ureg)
        assert_array_equal(result.data.magnitude, orig.data)
        assert result.data.units == ureg.Unit("m")

    def test_attach_units_from_attrs(self, example_unitless_da):
        orig = example_unitless_da
        result = orig.pint.quantify()
        assert_array_equal(result.data.magnitude, orig.data)
        assert str(result.data.units) == "meter"

        remaining_attrs = conversion.extract_unit_attributes(result)
        assert {k: v for k, v in remaining_attrs.items() if v is not None} == {}

    def test_attach_units_from_str_attr_no_unit(self, example_unitless_da):
        orig = example_unitless_da
        orig.attrs["units"] = "none"
        result = orig.pint.quantify("m")
        assert_array_equal(result.data.magnitude, orig.data)
        assert str(result.data.units) == "meter"

    def test_attach_units_given_unit_objs(self, example_unitless_da):
        orig = example_unitless_da
        ureg = UnitRegistry(force_ndarray=True)
        result = orig.pint.quantify(ureg.Unit("m"), unit_registry=ureg)
        assert_array_equal(result.data.magnitude, orig.data)
        assert result.data.units == ureg.Unit("m")

    @pytest.mark.parametrize("no_unit_value", conversion.no_unit_values)
    def test_override_units(self, example_unitless_da, no_unit_value):
        orig = example_unitless_da
        result = orig.pint.quantify(no_unit_value, u=no_unit_value)

        with pytest.raises(AttributeError):
            result.data.units
        with pytest.raises(AttributeError):
            result["u"].data.units

    def test_error_when_changing_units(self, example_quantity_da):
        da = example_quantity_da
        with pytest.raises(ValueError, match="already has units"):
            da.pint.quantify("s")

    def test_attach_no_units(self):
        arr = xr.DataArray([1, 2, 3], dims="x")
        quantified = arr.pint.quantify()
        assert_identical(quantified, arr)
        assert_units_equal(quantified, arr)

    def test_attach_no_new_units(self):
        da = xr.DataArray(unit_registry.Quantity([1, 2, 3], "m"), dims="x")
        quantified = da.pint.quantify()
        assert_identical(quantified, da)
        assert_units_equal(quantified, da)

    def test_attach_same_units(self):
        da = xr.DataArray(unit_registry.Quantity([1, 2, 3], "m"), dims="x")
        quantified = da.pint.quantify("m")
        assert_identical(quantified, da)
        assert_units_equal(quantified, da)

    def test_error_when_changing_units_dimension_coordinates(self):
        arr = xr.DataArray(
            [1, 2, 3],
            dims="x",
            coords={"x": ("x", [-1, 0, 1], {"units": unit_registry.Unit("m")})},
        )
        with pytest.raises(ValueError, match="already has units"):
            arr.pint.quantify({"x": "s"})

    def test_dimension_coordinate_array(self):
        ds = xr.Dataset(coords={"x": ("x", [10], {"units": "m"})})
        arr = ds.x

        # does not actually quantify because `arr` wraps a IndexVariable
        # but we still get a `Unit` in the attrs
        q = arr.pint.quantify()
        assert isinstance(q.attrs["units"], Unit)

    def test_dimension_coordinate_array_already_quantified(self):
        ds = xr.Dataset(coords={"x": ("x", [10], {"units": unit_registry.Unit("m")})})
        arr = ds.x

        with pytest.raises(ValueError):
            arr.pint.quantify({"x": "s"})

    def test_dimension_coordinate_array_already_quantified_same_units(self):
        ds = xr.Dataset(coords={"x": ("x", [10], {"units": unit_registry.Unit("m")})})
        arr = ds.x

        quantified = arr.pint.quantify({"x": "m"})

        assert_identical(quantified, arr)
        assert_units_equal(quantified, arr)

    def test_error_on_nonsense_units(self, example_unitless_da):
        da = example_unitless_da
        with pytest.raises(ValueError, match=str(da.name)):
            da.pint.quantify(units="aecjhbav")

    def test_error_on_nonsense_units_attrs(self, example_unitless_da):
        da = example_unitless_da
        da.attrs["units"] = "aecjhbav"
        with pytest.raises(
            ValueError, match=rf"{da.name}: {da.attrs['units']} \(attribute\)"
        ):
            da.pint.quantify()

    def test_parse_integer_inverse(self):
        # Regression test for issue #40
        da = xr.DataArray([10], attrs={"units": "m^-1"})
        result = da.pint.quantify()
        assert result.pint.units == Unit("1 / meter")


@pytest.mark.parametrize("formatter", ("", "P", "C"))
@pytest.mark.parametrize("modifier", ("", "~"))
def test_units_to_str_or_none(formatter, modifier):
    unit_format = f"{{:{modifier}{formatter}}}"
    unit_attrs = {None: "m", "a": "s", "b": "degC", "c": "degF", "d": "degK"}
    units = {key: unit_registry.Unit(value) for key, value in unit_attrs.items()}

    expected = {key: unit_format.format(value) for key, value in units.items()}
    actual = accessors.units_to_str_or_none(units, unit_format)

    assert expected == actual
    assert units == {key: unit_registry.Unit(value) for key, value in actual.items()}

    expected = {None: None}
    assert expected == accessors.units_to_str_or_none(expected, unit_format)


class TestDequantifyDataArray:
    def test_strip_units(self, example_quantity_da):
        result = example_quantity_da.pint.dequantify()
        assert isinstance(result.data, np.ndarray)
        assert isinstance(result.coords["x"].data, np.ndarray)

    def test_attrs_reinstated(self, example_quantity_da):
        da = example_quantity_da
        result = da.pint.dequantify()

        units = conversion.extract_units(da)
        attrs = conversion.extract_unit_attributes(result)

        assert units == attrs
        assert_all_str_or_none(attrs)

    def test_roundtrip_data(self, example_unitless_da):
        orig = example_unitless_da
        quantified = orig.pint.quantify()
        result = quantified.pint.dequantify()
        assert_equal(result, orig)

    def test_multiindex(self):
        mindex = pd.MultiIndex.from_product([["a", "b"], [1, 2]], names=("lat", "lon"))

        da = xr.DataArray(
            np.arange(len(mindex)), dims="multi", coords={"multi": mindex}
        )
        result = da.pint.dequantify()

        xr.testing.assert_identical(da, result)
        assert isinstance(result.indexes["multi"], pd.MultiIndex)


class TestPropertiesDataArray:
    def test_magnitude_getattr(self, example_quantity_da):
        da = example_quantity_da
        actual = da.pint.magnitude
        assert not isinstance(actual, Quantity)

    def test_magnitude_getattr_unitless(self, example_unitless_da):
        da = example_unitless_da
        xr.testing.assert_duckarray_equal(da.pint.magnitude, da.data)

    def test_units_getattr(self, example_quantity_da):
        da = example_quantity_da
        actual = da.pint.units
        assert isinstance(actual, Unit)
        assert actual == unit_registry.m

    def test_units_setattr(self, example_quantity_da):
        da = example_quantity_da
        with pytest.raises(ValueError):
            da.pint.units = "s"

    def test_units_getattr_unitless(self, example_unitless_da):
        da = example_unitless_da
        assert da.pint.units is None

    def test_units_setattr_unitless(self, example_unitless_da):
        da = example_unitless_da
        da.pint.units = unit_registry.s
        assert da.pint.units == unit_registry.s


@pytest.fixture()
def example_unitless_ds():
    users = np.linspace(0, 10, 20)
    funds = np.logspace(0, 10, 20)
    t = np.arange(20)
    ds = xr.Dataset(
        data_vars={"users": (["t"], users), "funds": (["t"], funds)}, coords={"t": t}
    )
    ds["users"].attrs["units"] = ""
    ds["funds"].attrs["units"] = "pound"
    return ds


@pytest.fixture()
def example_quantity_ds():
    users = np.linspace(0, 10, 20) * unit_registry.dimensionless
    funds = np.logspace(0, 10, 20) * unit_registry.pound
    t = np.arange(20)
    ds = xr.Dataset(
        data_vars={"users": (["t"], users), "funds": (["t"], funds)}, coords={"t": t}
    )
    return ds


class TestQuantifyDataSet:
    def test_attach_units_from_str(self, example_unitless_ds):
        orig = example_unitless_ds
        result = orig.pint.quantify()
        assert_array_equal(result["users"].data.magnitude, orig["users"].data)
        assert str(result["users"].data.units) == "dimensionless"

    def test_attach_units_given_registry(self, example_unitless_ds):
        orig = example_unitless_ds
        orig["users"].attrs.clear()
        result = orig.pint.quantify(
            {"users": "dimensionless"}, unit_registry=unit_registry
        )
        assert_array_equal(result["users"].data.magnitude, orig["users"].data)
        assert str(result["users"].data.units) == "dimensionless"

    def test_attach_units_from_attrs(self, example_unitless_ds):
        orig = example_unitless_ds
        orig["users"].attrs.clear()
        result = orig.pint.quantify({"users": "dimensionless"})
        assert_array_equal(result["users"].data.magnitude, orig["users"].data)
        assert str(result["users"].data.units) == "dimensionless"

        remaining_attrs = conversion.extract_unit_attributes(result)
        assert {k: v for k, v in remaining_attrs.items() if v is not None} == {}

    def test_attach_units_given_unit_objs(self, example_unitless_ds):
        orig = example_unitless_ds
        orig["users"].attrs.clear()
        dimensionless = unit_registry.Unit("dimensionless")
        result = orig.pint.quantify({"users": dimensionless})
        assert_array_equal(result["users"].data.magnitude, orig["users"].data)
        assert str(result["users"].data.units) == "dimensionless"

    def test_attach_units_from_str_attr_no_unit(self, example_unitless_ds):
        orig = example_unitless_ds
        orig["users"].attrs["units"] = "none"
        result = orig.pint.quantify({"users": "m"})
        assert_array_equal(result["users"].data.magnitude, orig["users"].data)
        assert str(result["users"].data.units) == "meter"

    @pytest.mark.parametrize("no_unit_value", conversion.no_unit_values)
    def test_override_units(self, example_unitless_ds, no_unit_value):
        orig = example_unitless_ds
        result = orig.pint.quantify({"users": no_unit_value})
        assert (
            getattr(result["users"].data, "units", "not a quantity") == "not a quantity"
        )

    def test_error_when_already_units(self, example_quantity_ds):
        with pytest.raises(ValueError, match="already has units"):
            example_quantity_ds.pint.quantify({"funds": "kg"})

    def test_attach_no_units(self):
        ds = xr.Dataset({"a": ("x", [1, 2, 3])})
        quantified = ds.pint.quantify()
        assert_identical(quantified, ds)
        assert_units_equal(quantified, ds)

    def test_attach_no_new_units(self):
        ds = xr.Dataset({"a": ("x", unit_registry.Quantity([1, 2, 3], "m"))})
        quantified = ds.pint.quantify()

        assert_identical(quantified, ds)
        assert_units_equal(quantified, ds)

    def test_attach_same_units(self):
        ds = xr.Dataset({"a": ("x", unit_registry.Quantity([1, 2, 3], "m"))})
        quantified = ds.pint.quantify({"a": "m"})

        assert_identical(quantified, ds)
        assert_units_equal(quantified, ds)

    def test_error_when_changing_units_dimension_coordinates(self):
        ds = xr.Dataset(
            coords={"x": ("x", [-1, 0, 1], {"units": unit_registry.Unit("m")})},
        )
        with pytest.raises(ValueError, match="already has units"):
            ds.pint.quantify({"x": "s"})

    def test_error_on_nonsense_units(self, example_unitless_ds):
        ds = example_unitless_ds
        with pytest.raises(ValueError):
            ds.pint.quantify(units={"users": "aecjhbav"})

    def test_error_on_nonsense_units_attrs(self, example_unitless_ds):
        ds = example_unitless_ds
        ds.users.attrs["units"] = "aecjhbav"
        with pytest.raises(
            ValueError, match=rf"'users': {ds.users.attrs['units']} \(attribute\)"
        ):
            ds.pint.quantify()

    def test_error_indicates_problematic_variable(self, example_unitless_ds):
        ds = example_unitless_ds
        with pytest.raises(ValueError, match="'users'"):
            ds.pint.quantify(units={"users": "aecjhbav"})

    def test_existing_units(self, example_quantity_ds):
        ds = example_quantity_ds.copy()
        ds.t.attrs["units"] = unit_registry.Unit("m")

        with pytest.raises(ValueError, match="Cannot attach"):
            ds.pint.quantify({"funds": "kg"})

    def test_existing_units_dimension(self, example_quantity_ds):
        ds = example_quantity_ds.copy()
        ds.t.attrs["units"] = unit_registry.Unit("m")

        with pytest.raises(ValueError, match="Cannot attach"):
            ds.pint.quantify({"t": "s"})


class TestDequantifyDataSet:
    def test_strip_units(self, example_quantity_ds):
        result = example_quantity_ds.pint.dequantify()

        assert all(
            isinstance(var.data, np.ndarray) for var in result.variables.values()
        )

    def test_attrs_reinstated(self, example_quantity_ds):
        ds = example_quantity_ds
        result = ds.pint.dequantify()

        units = conversion.extract_units(ds)
        # workaround for Unit("dimensionless") != str(Unit("dimensionless"))
        units = {
            key: str(value) if isinstance(value, Unit) else value
            for key, value in units.items()
        }

        attrs = conversion.extract_unit_attributes(result)

        assert units == attrs
        assert_all_str_or_none(attrs)

    def test_roundtrip_data(self, example_unitless_ds):
        orig = example_unitless_ds
        quantified = orig.pint.quantify()

        result = quantified.pint.dequantify()
        assert_equal(result, orig)

        result = quantified.pint.dequantify().pint.quantify()
        assert_equal(quantified, result)


@pytest.mark.parametrize(
    ["obj", "units", "expected", "error"],
    (
        pytest.param(
            xr.Dataset(
                {"a": ("x", Quantity([0, 1], "m")), "b": ("x", Quantity([2, 4], "s"))}
            ),
            {"a": "mm", "b": "ms"},
            xr.Dataset(
                {
                    "a": ("x", Quantity([0, 1000], "mm")),
                    "b": ("x", Quantity([2000, 4000], "ms")),
                }
            ),
            None,
            id="Dataset-compatible units-data",
        ),
        pytest.param(
            xr.Dataset(
                {"a": ("x", Quantity([0, 1], "km")), "b": ("x", Quantity([2, 4], "cm"))}
            ),
            "m",
            xr.Dataset(
                {
                    "a": ("x", Quantity([0, 1000], "m")),
                    "b": ("x", Quantity([0.02, 0.04], "m")),
                }
            ),
            None,
            id="Dataset-compatible units-data-str",
        ),
        pytest.param(
            xr.Dataset(
                {"a": ("x", Quantity([0, 1], "m")), "b": ("x", Quantity([2, 4], "s"))}
            ),
            {"a": "ms", "b": "mm"},
            None,
            ValueError,
            id="Dataset-incompatible units-data",
        ),
        pytest.param(
            xr.Dataset(coords={"x": ("x", [2, 4], {"units": Unit("s")})}),
            {"x": "ms"},
            xr.Dataset(coords={"x": ("x", [2000, 4000], {"units": Unit("ms")})}),
            None,
            id="Dataset-compatible units-dims",
        ),
        pytest.param(
            xr.Dataset(coords={"x": ("x", [2, 4], {"units": Unit("s")})}),
            {"x": "mm"},
            None,
            ValueError,
            id="Dataset-incompatible units-dims",
        ),
        pytest.param(
            xr.DataArray(Quantity([0, 1], "m"), dims="x"),
            {None: "mm"},
            xr.DataArray(Quantity([0, 1000], "mm"), dims="x"),
            None,
            id="DataArray-compatible units-data",
        ),
        pytest.param(
            xr.DataArray(Quantity([0, 1], "m"), dims="x"),
            "mm",
            xr.DataArray(Quantity([0, 1000], "mm"), dims="x"),
            None,
            id="DataArray-compatible units-data-str",
        ),
        pytest.param(
            xr.DataArray(Quantity([0, 1], "m"), dims="x", name="a"),
            {"a": "mm"},
            xr.DataArray(Quantity([0, 1000], "mm"), dims="x", name="a"),
            None,
            id="DataArray-compatible units-data-by name",
        ),
        pytest.param(
            xr.DataArray(Quantity([0, 1], "m"), dims="x"),
            {None: "ms"},
            None,
            ValueError,
            id="DataArray-incompatible units-data",
        ),
        pytest.param(
            xr.DataArray(
                [0, 1], dims="x", coords={"x": ("x", [2, 4], {"units": Unit("s")})}
            ),
            {"x": "ms"},
            xr.DataArray(
                [0, 1],
                dims="x",
                coords={"x": ("x", [2000, 4000], {"units": Unit("ms")})},
            ),
            None,
            id="DataArray-compatible units-dims",
        ),
        pytest.param(
            xr.DataArray(
                [0, 1], dims="x", coords={"x": ("x", [2, 4], {"units": Unit("s")})}
            ),
            {"x": "mm"},
            None,
            ValueError,
            id="DataArray-incompatible units-dims",
        ),
    ),
)
def test_to(obj, units, expected, error):
    if error is not None:
        with pytest.raises(error):
            obj.pint.to(units)
    else:
        actual = obj.pint.to(units)

        assert_units_equal(actual, expected)
        assert_identical(actual, expected)


@pytest.mark.parametrize(
    ["obj", "indexers", "expected", "error"],
    (
        pytest.param(
            xr.Dataset(
                {
                    "x": ("x", [10, 20, 30], {"units": unit_registry.Unit("dm")}),
                    "y": ("y", [60, 120], {"units": unit_registry.Unit("s")}),
                }
            ),
            {"x": Quantity([10, 30], "dm"), "y": Quantity([60], "s")},
            xr.Dataset(
                {
                    "x": ("x", [10, 30], {"units": unit_registry.Unit("dm")}),
                    "y": ("y", [60], {"units": unit_registry.Unit("s")}),
                }
            ),
            None,
            id="Dataset-identical units",
        ),
        pytest.param(
            xr.Dataset(
                {
                    "x": ("x", [10, 20, 30], {"units": unit_registry.Unit("dm")}),
                    "y": ("y", [60, 120], {"units": unit_registry.Unit("s")}),
                }
            ),
            {"x": Quantity([1, 3], "m"), "y": Quantity([1], "min")},
            xr.Dataset(
                {
                    "x": ("x", [1, 3], {"units": unit_registry.Unit("m")}),
                    "y": ("y", [1], {"units": unit_registry.Unit("min")}),
                }
            ),
            None,
            id="Dataset-compatible units",
        ),
        pytest.param(
            xr.Dataset(
                {
                    "x": ("x", [10, 20, 30], {"units": unit_registry.Unit("dm")}),
                    "y": ("y", [60, 120], {"units": unit_registry.Unit("s")}),
                }
            ),
            {"x": Quantity([1, 3], "s"), "y": Quantity([1], "m")},
            None,
            KeyError,
            id="Dataset-incompatible units",
        ),
        pytest.param(
            xr.DataArray(
                [[0, 1], [2, 3], [4, 5]],
                dims=("x", "y"),
                coords={
                    "x": ("x", [10, 20, 30], {"units": unit_registry.Unit("dm")}),
                    "y": ("y", [60, 120], {"units": unit_registry.Unit("s")}),
                },
            ),
            {"x": Quantity([10, 30], "dm"), "y": Quantity([60], "s")},
            xr.DataArray(
                [[0], [4]],
                dims=("x", "y"),
                coords={
                    "x": ("x", [10, 30], {"units": unit_registry.Unit("dm")}),
                    "y": ("y", [60], {"units": unit_registry.Unit("s")}),
                },
            ),
            None,
            id="DataArray-identical units",
        ),
        pytest.param(
            xr.DataArray(
                [[0, 1], [2, 3], [4, 5]],
                dims=("x", "y"),
                coords={
                    "x": ("x", [10, 20, 30], {"units": unit_registry.Unit("dm")}),
                    "y": ("y", [60, 120], {"units": unit_registry.Unit("s")}),
                },
            ),
            {"x": Quantity([1, 3], "m"), "y": Quantity([1], "min")},
            xr.DataArray(
                [[0], [4]],
                dims=("x", "y"),
                coords={
                    "x": ("x", [1, 3], {"units": unit_registry.Unit("m")}),
                    "y": ("y", [1], {"units": unit_registry.Unit("min")}),
                },
            ),
            None,
            id="DataArray-compatible units",
        ),
        pytest.param(
            xr.DataArray(
                [[0, 1], [2, 3], [4, 5]],
                dims=("x", "y"),
                coords={
                    "x": ("x", [10, 20, 30], {"units": unit_registry.Unit("dm")}),
                    "y": ("y", [60, 120], {"units": unit_registry.Unit("s")}),
                },
            ),
            {"x": Quantity([10, 30], "s"), "y": Quantity([60], "m")},
            None,
            KeyError,
            id="DataArray-incompatible units",
        ),
    ),
)
def test_sel(obj, indexers, expected, error):
    if error is not None:
        with pytest.raises(error):
            obj.pint.sel(indexers)
    else:
        actual = obj.pint.sel(indexers)
        assert_units_equal(actual, expected)
        assert_identical(actual, expected)


@pytest.mark.parametrize(
    ["obj", "indexers", "expected", "error"],
    (
        pytest.param(
            xr.Dataset(
                {
                    "x": ("x", [10, 20, 30], {"units": unit_registry.Unit("dm")}),
                    "y": ("y", [60, 120], {"units": unit_registry.Unit("s")}),
                }
            ),
            {"x": Quantity([10, 30], "dm"), "y": Quantity([60], "s")},
            xr.Dataset(
                {
                    "x": ("x", [10, 30], {"units": unit_registry.Unit("dm")}),
                    "y": ("y", [60], {"units": unit_registry.Unit("s")}),
                }
            ),
            None,
            id="Dataset-identical units",
        ),
        pytest.param(
            xr.Dataset(
                {
                    "x": ("x", [10, 20, 30], {"units": unit_registry.Unit("dm")}),
                    "y": ("y", [60, 120], {"units": unit_registry.Unit("s")}),
                }
            ),
            {"x": Quantity([1, 3], "m"), "y": Quantity([1], "min")},
            xr.Dataset(
                {
                    "x": ("x", [1, 3], {"units": unit_registry.Unit("m")}),
                    "y": ("y", [1], {"units": unit_registry.Unit("min")}),
                }
            ),
            None,
            id="Dataset-compatible units",
        ),
        pytest.param(
            xr.Dataset(
                {
                    "x": ("x", [10, 20, 30], {"units": unit_registry.Unit("dm")}),
                    "y": ("y", [60, 120], {"units": unit_registry.Unit("s")}),
                }
            ),
            {"x": Quantity([1, 3], "s"), "y": Quantity([1], "m")},
            None,
            KeyError,
            id="Dataset-incompatible units",
        ),
        pytest.param(
            xr.DataArray(
                [[0, 1], [2, 3], [4, 5]],
                dims=("x", "y"),
                coords={
                    "x": ("x", [10, 20, 30], {"units": unit_registry.Unit("dm")}),
                    "y": ("y", [60, 120], {"units": unit_registry.Unit("s")}),
                },
            ),
            {"x": Quantity([10, 30], "dm"), "y": Quantity([60], "s")},
            xr.DataArray(
                [[0], [4]],
                dims=("x", "y"),
                coords={
                    "x": ("x", [10, 30], {"units": unit_registry.Unit("dm")}),
                    "y": ("y", [60], {"units": unit_registry.Unit("s")}),
                },
            ),
            None,
            id="DataArray-identical units",
        ),
        pytest.param(
            xr.DataArray(
                [[0, 1], [2, 3], [4, 5]],
                dims=("x", "y"),
                coords={
                    "x": ("x", [10, 20, 30], {"units": unit_registry.Unit("dm")}),
                    "y": ("y", [60, 120], {"units": unit_registry.Unit("s")}),
                },
            ),
            {"x": Quantity([1, 3], "m"), "y": Quantity([1], "min")},
            xr.DataArray(
                [[0], [4]],
                dims=("x", "y"),
                coords={
                    "x": ("x", [1, 3], {"units": unit_registry.Unit("m")}),
                    "y": ("y", [1], {"units": unit_registry.Unit("min")}),
                },
            ),
            None,
            id="DataArray-compatible units",
        ),
        pytest.param(
            xr.DataArray(
                [[0, 1], [2, 3], [4, 5]],
                dims=("x", "y"),
                coords={
                    "x": ("x", [10, 20, 30], {"units": unit_registry.Unit("dm")}),
                    "y": ("y", [60, 120], {"units": unit_registry.Unit("s")}),
                },
            ),
            {"x": Quantity([10, 30], "s"), "y": Quantity([60], "m")},
            None,
            KeyError,
            id="DataArray-incompatible units",
        ),
    ),
)
def test_loc(obj, indexers, expected, error):
    if error is not None:
        with pytest.raises(error):
            obj.pint.loc[indexers]
    else:
        actual = obj.pint.loc[indexers]
        assert_units_equal(actual, expected)
        assert_identical(actual, expected)


@pytest.mark.parametrize(
    ["obj", "indexers", "values", "expected", "error"],
    (
        pytest.param(
            xr.DataArray(
                [[0, 1], [2, 3], [4, 5]],
                dims=("x", "y"),
                coords={
                    "x": ("x", [10, 20, 30], {"units": unit_registry.Unit("dm")}),
                    "y": ("y", [60, 120], {"units": unit_registry.Unit("s")}),
                },
            ),
            {"x": Quantity([10, 30], "dm"), "y": Quantity([60], "s")},
            [[-1], [-2]],
            xr.DataArray(
                [[-1, 1], [2, 3], [-2, 5]],
                dims=("x", "y"),
                coords={
                    "x": ("x", [10, 20, 30], {"units": unit_registry.Unit("dm")}),
                    "y": ("y", [60, 120], {"units": unit_registry.Unit("s")}),
                },
            ),
            None,
            id="coords-identical units",
        ),
        pytest.param(
            xr.DataArray(
                [[0, 1], [2, 3], [4, 5]],
                dims=("x", "y"),
                coords={
                    "x": ("x", [10, 20, 30], {"units": unit_registry.Unit("dm")}),
                    "y": ("y", [60, 120], {"units": unit_registry.Unit("s")}),
                },
            ),
            {"x": Quantity([1, 3], "m"), "y": Quantity([1], "min")},
            [[-1], [-2]],
            xr.DataArray(
                [[-1, 1], [2, 3], [-2, 5]],
                dims=("x", "y"),
                coords={
                    "x": ("x", [10, 20, 30], {"units": unit_registry.Unit("dm")}),
                    "y": ("y", [60, 120], {"units": unit_registry.Unit("s")}),
                },
            ),
            None,
            id="coords-compatible units",
        ),
        pytest.param(
            xr.DataArray(
                [[0, 1], [2, 3], [4, 5]],
                dims=("x", "y"),
                coords={
                    "x": ("x", [10, 20, 30], {"units": unit_registry.Unit("dm")}),
                    "y": ("y", [60, 120], {"units": unit_registry.Unit("s")}),
                },
            ),
            {"x": Quantity([1, 3], "s"), "y": Quantity([1], "m")},
            [[-1], [-2]],
            None,
            KeyError,
            id="coords-incompatible units",
        ),
        pytest.param(
            xr.DataArray(
                Quantity([[0, 1], [2, 3], [4, 5]], "m"),
                dims=("x", "y"),
                coords={
                    "x": ("x", [10, 20, 30], {"units": unit_registry.Unit("dm")}),
                    "y": ("y", [60, 120], {"units": unit_registry.Unit("s")}),
                },
            ),
            {"x": Quantity([10, 30], "dm"), "y": Quantity([60], "s")},
            Quantity([[-1], [-2]], "m"),
            xr.DataArray(
                Quantity([[-1, 1], [2, 3], [-2, 5]], "m"),
                dims=("x", "y"),
                coords={
                    "x": ("x", [10, 20, 30], {"units": unit_registry.Unit("dm")}),
                    "y": ("y", [60, 120], {"units": unit_registry.Unit("s")}),
                },
            ),
            None,
            id="data-identical units",
        ),
        pytest.param(
            xr.DataArray(
                Quantity([[0, 1], [2, 3], [4, 5]], "m"),
                dims=("x", "y"),
                coords={
                    "x": ("x", [10, 20, 30], {"units": unit_registry.Unit("dm")}),
                    "y": ("y", [60, 120], {"units": unit_registry.Unit("s")}),
                },
            ),
            {"x": Quantity([10, 30], "dm"), "y": Quantity([60], "s")},
            Quantity([[-1], [-2]], "km"),
            xr.DataArray(
                Quantity([[-1000, 1], [2, 3], [-2000, 5]], "m"),
                dims=("x", "y"),
                coords={
                    "x": ("x", [10, 20, 30], {"units": unit_registry.Unit("dm")}),
                    "y": ("y", [60, 120], {"units": unit_registry.Unit("s")}),
                },
            ),
            None,
            id="data-compatible units",
        ),
        pytest.param(
            xr.DataArray(
                Quantity([[0, 1], [2, 3], [4, 5]], "m"),
                dims=("x", "y"),
                coords={
                    "x": ("x", [10, 20, 30], {"units": unit_registry.Unit("dm")}),
                    "y": ("y", [60, 120], {"units": unit_registry.Unit("s")}),
                },
            ),
            {"x": Quantity([10, 30], "dm"), "y": Quantity([60], "s")},
            Quantity([[-1], [-2]], "s"),
            None,
            pint.DimensionalityError,
            id="data-incompatible units",
        ),
    ),
)
def test_loc_setitem(obj, indexers, values, expected, error):
    if error is not None:
        with pytest.raises(error):
            obj.pint.loc[indexers] = values
    else:
        obj.pint.loc[indexers] = values
        assert_units_equal(obj, expected)
        assert_identical(obj, expected)


@pytest.mark.parametrize(
    ["obj", "indexers", "expected", "error"],
    (
        pytest.param(
            xr.Dataset(
                {
                    "x": ("x", [10, 20, 30], {"units": unit_registry.Unit("dm")}),
                    "y": ("y", [60, 120], {"units": unit_registry.Unit("s")}),
                }
            ),
            {"x": Quantity([10, 30], "dm"), "y": Quantity([60], "s")},
            xr.Dataset(
                {
                    "x": ("x", [20], {"units": unit_registry.Unit("dm")}),
                    "y": ("y", [120], {"units": unit_registry.Unit("s")}),
                }
            ),
            None,
            id="Dataset-identical units",
        ),
        pytest.param(
            xr.Dataset(
                {
                    "x": ("x", [10, 20, 30], {"units": unit_registry.Unit("dm")}),
                    "y": ("y", [60, 120], {"units": unit_registry.Unit("s")}),
                }
            ),
            {"x": Quantity([1, 3], "m"), "y": Quantity([1], "min")},
            xr.Dataset(
                {
                    "x": ("x", [20], {"units": unit_registry.Unit("dm")}),
                    "y": ("y", [120], {"units": unit_registry.Unit("s")}),
                }
            ),
            None,
            id="Dataset-compatible units",
        ),
        pytest.param(
            xr.Dataset(
                {
                    "x": ("x", [10, 20, 30], {"units": unit_registry.Unit("dm")}),
                    "y": ("y", [60, 120], {"units": unit_registry.Unit("s")}),
                }
            ),
            {"x": Quantity([1, 3], "s"), "y": Quantity([1], "m")},
            None,
            KeyError,
            id="Dataset-incompatible units",
        ),
        pytest.param(
            xr.Dataset(
                {
                    "x": ("x", [10, 20, 30], {"units": unit_registry.Unit("dm")}),
                    "y": ("y", [60, 120], {"units": unit_registry.Unit("s")}),
                }
            ),
            {"x": Quantity([10, 30], "m"), "y": Quantity([60], "min")},
            None,
            KeyError,
            id="Dataset-compatible units-not found",
        ),
        pytest.param(
            xr.DataArray(
                [[0, 1], [2, 3], [4, 5]],
                dims=("x", "y"),
                coords={
                    "x": ("x", [10, 20, 30], {"units": unit_registry.Unit("dm")}),
                    "y": ("y", [60, 120], {"units": unit_registry.Unit("s")}),
                },
            ),
            {"x": Quantity([10, 30], "dm"), "y": Quantity([60], "s")},
            xr.DataArray(
                [[3]],
                dims=("x", "y"),
                coords={
                    "x": ("x", [20], {"units": unit_registry.Unit("dm")}),
                    "y": ("y", [120], {"units": unit_registry.Unit("s")}),
                },
            ),
            None,
            id="DataArray-identical units",
        ),
        pytest.param(
            xr.DataArray(
                [[0, 1], [2, 3], [4, 5]],
                dims=("x", "y"),
                coords={
                    "x": ("x", [10, 20, 30], {"units": unit_registry.Unit("dm")}),
                    "y": ("y", [60, 120], {"units": unit_registry.Unit("s")}),
                },
            ),
            {"x": Quantity([1, 3], "m"), "y": Quantity([1], "min")},
            xr.DataArray(
                [[3]],
                dims=("x", "y"),
                coords={
                    "x": ("x", [20], {"units": unit_registry.Unit("dm")}),
                    "y": ("y", [120], {"units": unit_registry.Unit("s")}),
                },
            ),
            None,
            id="DataArray-compatible units",
        ),
        pytest.param(
            xr.DataArray(
                [[0, 1], [2, 3], [4, 5]],
                dims=("x", "y"),
                coords={
                    "x": ("x", [10, 20, 30], {"units": unit_registry.Unit("dm")}),
                    "y": ("y", [60, 120], {"units": unit_registry.Unit("s")}),
                },
            ),
            {"x": Quantity([10, 30], "s"), "y": Quantity([60], "m")},
            None,
            KeyError,
            id="DataArray-incompatible units",
        ),
        pytest.param(
            xr.DataArray(
                [[0, 1], [2, 3], [4, 5]],
                dims=("x", "y"),
                coords={
                    "x": ("x", [10, 20, 30], {"units": unit_registry.Unit("dm")}),
                    "y": ("y", [60, 120], {"units": unit_registry.Unit("s")}),
                },
            ),
            {"x": Quantity([10, 30], "m"), "y": Quantity([60], "min")},
            None,
            KeyError,
            id="DataArray-compatible units-not found",
        ),
    ),
)
def test_drop_sel(obj, indexers, expected, error):
    if error is not None:
        with pytest.raises(error):
            obj.pint.drop_sel(indexers)
    else:
        actual = obj.pint.drop_sel(indexers)
        assert_units_equal(actual, expected)
        assert_identical(actual, expected)


@requires_dask_array
@pytest.mark.parametrize(
    "obj",
    (
        pytest.param(
            xr.Dataset(
                {"a": ("x", np.linspace(0, 1, 11))},
                coords={"u": ("x", np.arange(11))},
            ),
            id="Dataset-no units",
        ),
        pytest.param(
            xr.Dataset(
                {
                    "a": (
                        "x",
                        Quantity(np.linspace(0, 1, 11), "m"),
                    )
                },
                coords={
                    "u": (
                        "x",
                        Quantity(np.arange(11), "m"),
                    )
                },
            ),
            id="Dataset-units",
        ),
        pytest.param(
            xr.DataArray(
                np.linspace(0, 1, 11),
                coords={
                    "u": (
                        "x",
                        np.arange(11),
                    )
                },
                dims="x",
            ),
            id="DataArray-no units",
        ),
        pytest.param(
            xr.DataArray(
                Quantity(np.linspace(0, 1, 11), "m"),
                coords={
                    "u": (
                        "x",
                        Quantity(np.arange(11), "m"),
                    )
                },
                dims="x",
            ),
            id="DataArray-units",
        ),
    ),
)
def test_chunk(obj):
    actual = obj.pint.chunk({"x": 2})

    expected = (
        obj.pint.dequantify().chunk({"x": 2}).pint.quantify(unit_registry=unit_registry)
    )

    assert_units_equal(actual, expected)
    assert_identical(actual, expected)


@pytest.mark.parametrize(
    ["obj", "indexers", "expected", "error"],
    (
        pytest.param(
            xr.Dataset(
                {
                    "x": ("x", [10, 20, 30], {"units": unit_registry.Unit("dm")}),
                    "y": ("y", [60, 120], {"units": unit_registry.Unit("s")}),
                }
            ),
            {"x": Quantity([10, 30, 50], "dm"), "y": Quantity([0, 120, 240], "s")},
            xr.Dataset(
                {
                    "x": ("x", [10, 30, 50], {"units": unit_registry.Unit("dm")}),
                    "y": ("y", [0, 120, 240], {"units": unit_registry.Unit("s")}),
                }
            ),
            None,
            id="Dataset-identical units",
        ),
        pytest.param(
            xr.Dataset(
                {
                    "x": ("x", [10, 20, 30], {"units": unit_registry.Unit("dm")}),
                    "y": ("y", [60, 120], {"units": unit_registry.Unit("s")}),
                }
            ),
            {"x": Quantity([0, 1, 3, 5], "m"), "y": Quantity([0, 2, 4], "min")},
            xr.Dataset(
                {
                    "x": ("x", [0, 1, 3, 5], {"units": unit_registry.Unit("m")}),
                    "y": ("y", [0, 2, 4], {"units": unit_registry.Unit("min")}),
                }
            ),
            None,
            id="Dataset-compatible units",
        ),
        pytest.param(
            xr.Dataset(
                {
                    "x": ("x", [10, 20, 30], {"units": unit_registry.Unit("dm")}),
                    "y": ("y", [60, 120], {"units": unit_registry.Unit("s")}),
                }
            ),
            {"x": Quantity([1, 3], "s"), "y": Quantity([1], "m")},
            None,
            ValueError,
            id="Dataset-incompatible units",
        ),
        pytest.param(
            xr.DataArray(
                [[0, 1], [2, 3], [4, 5]],
                dims=("x", "y"),
                coords={
                    "x": ("x", [10, 20, 30], {"units": unit_registry.Unit("dm")}),
                    "y": ("y", [60, 120], {"units": unit_registry.Unit("s")}),
                },
            ),
            {"x": Quantity([10, 30, 50], "dm"), "y": Quantity([0, 240], "s")},
            xr.DataArray(
                [[np.nan, np.nan], [np.nan, np.nan], [np.nan, np.nan]],
                dims=("x", "y"),
                coords={
                    "x": ("x", [10, 30, 50], {"units": unit_registry.Unit("dm")}),
                    "y": ("y", [0, 240], {"units": unit_registry.Unit("s")}),
                },
            ),
            None,
            id="DataArray-identical units",
        ),
        pytest.param(
            xr.DataArray(
                [[0, 1], [2, 3], [4, 5]],
                dims=("x", "y"),
                coords={
                    "x": ("x", [10, 20, 30], {"units": unit_registry.Unit("dm")}),
                    "y": ("y", [60, 120], {"units": unit_registry.Unit("s")}),
                },
            ),
            {"x": Quantity([1, 3, 5], "m"), "y": Quantity([0, 2], "min")},
            xr.DataArray(
                [[np.nan, 1], [np.nan, 5], [np.nan, np.nan]],
                dims=("x", "y"),
                coords={
                    "x": ("x", [1, 3, 5], {"units": unit_registry.Unit("m")}),
                    "y": ("y", [0, 2], {"units": unit_registry.Unit("min")}),
                },
            ),
            None,
            id="DataArray-compatible units",
        ),
        pytest.param(
            xr.DataArray(
                [[0, 1], [2, 3], [4, 5]],
                dims=("x", "y"),
                coords={
                    "x": ("x", [10, 20, 30], {"units": unit_registry.Unit("dm")}),
                    "y": ("y", [60, 120], {"units": unit_registry.Unit("s")}),
                },
            ),
            {"x": Quantity([10, 30], "s"), "y": Quantity([60], "m")},
            None,
            ValueError,
            id="DataArray-incompatible units",
        ),
    ),
)
def test_reindex(obj, indexers, expected, error):
    if error is not None:
        with pytest.raises(error):
            obj.pint.reindex(indexers)
    else:
        actual = obj.pint.reindex(indexers)
        assert_units_equal(actual, expected)
        assert_identical(actual, expected)


@pytest.mark.parametrize(
    ["obj", "other", "expected", "error"],
    (
        pytest.param(
            xr.Dataset(
                {
                    "x": ("x", [10, 20, 30], {"units": unit_registry.Unit("dm")}),
                    "y": ("y", [60, 120], {"units": unit_registry.Unit("s")}),
                }
            ),
            xr.Dataset(
                {
                    "x": ("x", [10, 30, 50], {"units": unit_registry.Unit("dm")}),
                    "y": ("y", [0, 120, 240], {"units": unit_registry.Unit("s")}),
                }
            ),
            xr.Dataset(
                {
                    "x": ("x", [10, 30, 50], {"units": unit_registry.Unit("dm")}),
                    "y": ("y", [0, 120, 240], {"units": unit_registry.Unit("s")}),
                }
            ),
            None,
            id="Dataset-identical units",
        ),
        pytest.param(
            xr.Dataset(
                {
                    "x": ("x", [10, 20, 30], {"units": unit_registry.Unit("dm")}),
                    "y": ("y", [60, 120], {"units": unit_registry.Unit("s")}),
                }
            ),
            xr.Dataset(
                {
                    "x": ("x", [0, 1, 3, 5], {"units": unit_registry.Unit("m")}),
                    "y": ("y", [0, 2, 4], {"units": unit_registry.Unit("min")}),
                }
            ),
            xr.Dataset(
                {
                    "x": ("x", [0, 1, 3, 5], {"units": unit_registry.Unit("m")}),
                    "y": ("y", [0, 2, 4], {"units": unit_registry.Unit("min")}),
                }
            ),
            None,
            id="Dataset-compatible units",
        ),
        pytest.param(
            xr.Dataset(
                {
                    "x": ("x", [10, 20, 30], {"units": unit_registry.Unit("dm")}),
                    "y": ("y", [60, 120], {"units": unit_registry.Unit("s")}),
                }
            ),
            xr.Dataset(
                {
                    "x": ("x", [1, 3], {"units": unit_registry.Unit("s")}),
                    "y": ("y", [1], {"units": unit_registry.Unit("m")}),
                }
            ),
            None,
            ValueError,
            id="Dataset-incompatible units",
        ),
        pytest.param(
            xr.DataArray(
                [[0, 1], [2, 3], [4, 5]],
                dims=("x", "y"),
                coords={
                    "x": ("x", [10, 20, 30], {"units": unit_registry.Unit("dm")}),
                    "y": ("y", [60, 120], {"units": unit_registry.Unit("s")}),
                },
            ),
            xr.Dataset(
                {
                    "x": ("x", [10, 30, 50], {"units": unit_registry.Unit("dm")}),
                    "y": ("y", [0, 240], {"units": unit_registry.Unit("s")}),
                }
            ),
            xr.DataArray(
                [[np.nan, np.nan], [np.nan, np.nan], [np.nan, np.nan]],
                dims=("x", "y"),
                coords={
                    "x": ("x", [10, 30, 50], {"units": unit_registry.Unit("dm")}),
                    "y": ("y", [0, 240], {"units": unit_registry.Unit("s")}),
                },
            ),
            None,
            id="DataArray-identical units",
        ),
        pytest.param(
            xr.DataArray(
                [[0, 1], [2, 3], [4, 5]],
                dims=("x", "y"),
                coords={
                    "x": ("x", [10, 20, 30], {"units": unit_registry.Unit("dm")}),
                    "y": ("y", [60, 120], {"units": unit_registry.Unit("s")}),
                },
            ),
            xr.Dataset(
                {
                    "x": ("x", [1, 3, 5], {"units": unit_registry.Unit("m")}),
                    "y": ("y", [0, 2], {"units": unit_registry.Unit("min")}),
                }
            ),
            xr.DataArray(
                [[np.nan, 1], [np.nan, 5], [np.nan, np.nan]],
                dims=("x", "y"),
                coords={
                    "x": ("x", [1, 3, 5], {"units": unit_registry.Unit("m")}),
                    "y": ("y", [0, 2], {"units": unit_registry.Unit("min")}),
                },
            ),
            None,
            id="DataArray-compatible units",
        ),
        pytest.param(
            xr.DataArray(
                [[0, 1], [2, 3], [4, 5]],
                dims=("x", "y"),
                coords={
                    "x": ("x", [10, 20, 30], {"units": unit_registry.Unit("dm")}),
                    "y": ("y", [60, 120], {"units": unit_registry.Unit("s")}),
                },
            ),
            xr.Dataset(
                {
                    "x": ("x", [10, 30], {"units": unit_registry.Unit("s")}),
                    "y": ("y", [60], {"units": unit_registry.Unit("m")}),
                }
            ),
            None,
            ValueError,
            id="DataArray-incompatible units",
        ),
    ),
)
def test_reindex_like(obj, other, expected, error):
    if error is not None:
        with pytest.raises(error):
            obj.pint.reindex_like(other)
    else:
        actual = obj.pint.reindex_like(other)
        assert_units_equal(actual, expected)
        assert_identical(actual, expected)


@requires_scipy
@pytest.mark.parametrize(
    ["obj", "indexers", "expected", "error"],
    (
        pytest.param(
            xr.Dataset(
                {
                    "x": ("x", [10, 20, 30], {"units": unit_registry.Unit("dm")}),
                    "y": ("y", [60, 120], {"units": unit_registry.Unit("s")}),
                }
            ),
            {"x": Quantity([10, 30, 50], "dm"), "y": Quantity([0, 120, 240], "s")},
            xr.Dataset(
                {
                    "x": ("x", [10, 30, 50], {"units": unit_registry.Unit("dm")}),
                    "y": ("y", [0, 120, 240], {"units": unit_registry.Unit("s")}),
                }
            ),
            None,
            id="Dataset-identical units",
        ),
        pytest.param(
            xr.Dataset(
                {
                    "x": ("x", [10, 20, 30], {"units": unit_registry.Unit("dm")}),
                    "y": ("y", [60, 120], {"units": unit_registry.Unit("s")}),
                }
            ),
            {"x": Quantity([0, 1, 3, 5], "m"), "y": Quantity([0, 2, 4], "min")},
            xr.Dataset(
                {
                    "x": ("x", [0, 1, 3, 5], {"units": unit_registry.Unit("m")}),
                    "y": ("y", [0, 2, 4], {"units": unit_registry.Unit("min")}),
                }
            ),
            None,
            id="Dataset-compatible units",
        ),
        pytest.param(
            xr.Dataset(
                {
                    "x": ("x", [10, 20, 30], {"units": unit_registry.Unit("dm")}),
                    "y": ("y", [60, 120], {"units": unit_registry.Unit("s")}),
                }
            ),
            {"x": Quantity([1, 3], "s"), "y": Quantity([1], "m")},
            None,
            ValueError,
            id="Dataset-incompatible units",
        ),
        pytest.param(
            xr.Dataset(
                {
                    "a": (("x", "y"), Quantity([[0, 1], [2, 3], [4, 5]], "kg")),
                    "x": [10, 20, 30],
                    "y": [60, 120],
                }
            ),
            {
                "x": [15, 25],
                "y": [75, 105],
            },
            xr.Dataset(
                {
                    "a": (("x", "y"), Quantity([[1.25, 1.75], [3.25, 3.75]], "kg")),
                    "x": [15, 25],
                    "y": [75, 105],
                }
            ),
            None,
            id="Dataset-data units",
        ),
        pytest.param(
            xr.DataArray(
                [[0, 1], [2, 3], [4, 5]],
                dims=("x", "y"),
                coords={
                    "x": ("x", [10, 20, 30], {"units": unit_registry.Unit("dm")}),
                    "y": ("y", [60, 120], {"units": unit_registry.Unit("s")}),
                },
            ),
            {"x": Quantity([10, 30, 50], "dm"), "y": Quantity([0, 240], "s")},
            xr.DataArray(
                [[np.nan, np.nan], [np.nan, np.nan], [np.nan, np.nan]],
                dims=("x", "y"),
                coords={
                    "x": ("x", [10, 30, 50], {"units": unit_registry.Unit("dm")}),
                    "y": ("y", [0, 240], {"units": unit_registry.Unit("s")}),
                },
            ),
            None,
            id="DataArray-identical units",
        ),
        pytest.param(
            xr.DataArray(
                [[0, 1], [2, 3], [4, 5]],
                dims=("x", "y"),
                coords={
                    "x": ("x", [10, 20, 30], {"units": unit_registry.Unit("dm")}),
                    "y": ("y", [60, 120], {"units": unit_registry.Unit("s")}),
                },
            ),
            {"x": Quantity([1, 3, 5], "m"), "y": Quantity([0, 2], "min")},
            xr.DataArray(
                [[np.nan, 1], [np.nan, 5], [np.nan, np.nan]],
                dims=("x", "y"),
                coords={
                    "x": ("x", [1, 3, 5], {"units": unit_registry.Unit("m")}),
                    "y": ("y", [0, 2], {"units": unit_registry.Unit("min")}),
                },
            ),
            None,
            id="DataArray-compatible units",
        ),
        pytest.param(
            xr.DataArray(
                [[0, 1], [2, 3], [4, 5]],
                dims=("x", "y"),
                coords={
                    "x": ("x", [10, 20, 30], {"units": unit_registry.Unit("dm")}),
                    "y": ("y", [60, 120], {"units": unit_registry.Unit("s")}),
                },
            ),
            {"x": Quantity([10, 30], "s"), "y": Quantity([60], "m")},
            None,
            ValueError,
            id="DataArray-incompatible units",
        ),
        pytest.param(
            xr.DataArray(
                Quantity([[0, 1], [2, 3], [4, 5]], "kg"),
                dims=("x", "y"),
                coords={
                    "x": [10, 20, 30],
                    "y": [60, 120],
                },
            ),
            {
                "x": [15, 25],
                "y": [75, 105],
            },
            xr.DataArray(
                Quantity([[1.25, 1.75], [3.25, 3.75]], "kg"),
                dims=("x", "y"),
                coords={
                    "x": [15, 25],
                    "y": [75, 105],
                },
            ),
            None,
            id="DataArray-data units",
        ),
    ),
)
def test_interp(obj, indexers, expected, error):
    if error is not None:
        with pytest.raises(error):
            obj.pint.interp(indexers)
    else:
        actual = obj.pint.interp(indexers)
        assert_units_equal(actual, expected)
        assert_identical(actual, expected)


@requires_scipy
@pytest.mark.parametrize(
    ["obj", "other", "expected", "error"],
    (
        pytest.param(
            xr.Dataset(
                {
                    "x": ("x", [10, 20, 30], {"units": unit_registry.Unit("dm")}),
                    "y": ("y", [60, 120], {"units": unit_registry.Unit("s")}),
                }
            ),
            xr.Dataset(
                {
                    "x": ("x", [10, 30, 50], {"units": unit_registry.Unit("dm")}),
                    "y": ("y", [0, 120, 240], {"units": unit_registry.Unit("s")}),
                }
            ),
            xr.Dataset(
                {
                    "x": ("x", [10, 30, 50], {"units": unit_registry.Unit("dm")}),
                    "y": ("y", [0, 120, 240], {"units": unit_registry.Unit("s")}),
                }
            ),
            None,
            id="Dataset-identical units",
        ),
        pytest.param(
            xr.Dataset(
                {
                    "x": ("x", [10, 20, 30], {"units": unit_registry.Unit("dm")}),
                    "y": ("y", [60, 120], {"units": unit_registry.Unit("s")}),
                }
            ),
            xr.Dataset(
                {
                    "x": ("x", [0, 1, 3, 5], {"units": unit_registry.Unit("m")}),
                    "y": ("y", [0, 2, 4], {"units": unit_registry.Unit("min")}),
                }
            ),
            xr.Dataset(
                {
                    "x": ("x", [0, 1, 3, 5], {"units": unit_registry.Unit("m")}),
                    "y": ("y", [0, 2, 4], {"units": unit_registry.Unit("min")}),
                }
            ),
            None,
            id="Dataset-compatible units",
        ),
        pytest.param(
            xr.Dataset(
                {
                    "x": ("x", [10, 20, 30], {"units": unit_registry.Unit("dm")}),
                    "y": ("y", [60, 120], {"units": unit_registry.Unit("s")}),
                }
            ),
            xr.Dataset(
                {
                    "x": ("x", [1, 3], {"units": unit_registry.Unit("s")}),
                    "y": ("y", [1], {"units": unit_registry.Unit("m")}),
                }
            ),
            None,
            ValueError,
            id="Dataset-incompatible units",
        ),
        pytest.param(
            xr.DataArray(
                [[0, 1], [2, 3], [4, 5]],
                dims=("x", "y"),
                coords={
                    "x": ("x", [10, 20, 30], {"units": unit_registry.Unit("dm")}),
                    "y": ("y", [60, 120], {"units": unit_registry.Unit("s")}),
                },
            ),
            xr.Dataset(
                {
                    "x": ("x", [10, 30, 50], {"units": unit_registry.Unit("dm")}),
                    "y": ("y", [0, 240], {"units": unit_registry.Unit("s")}),
                }
            ),
            xr.DataArray(
                [[np.nan, np.nan], [np.nan, np.nan], [np.nan, np.nan]],
                dims=("x", "y"),
                coords={
                    "x": ("x", [10, 30, 50], {"units": unit_registry.Unit("dm")}),
                    "y": ("y", [0, 240], {"units": unit_registry.Unit("s")}),
                },
            ),
            None,
            id="DataArray-identical units",
        ),
        pytest.param(
            xr.Dataset(
                {
                    "a": (("x", "y"), Quantity([[0, 1], [2, 3], [4, 5]], "kg")),
                    "x": [10, 20, 30],
                    "y": [60, 120],
                }
            ),
            xr.Dataset(
                {
                    "x": [15, 25],
                    "y": [75, 105],
                }
            ),
            xr.Dataset(
                {
                    "a": (("x", "y"), Quantity([[1.25, 1.75], [3.25, 3.75]], "kg")),
                    "x": [15, 25],
                    "y": [75, 105],
                }
            ),
            None,
            id="Dataset-data units",
        ),
        pytest.param(
            xr.DataArray(
                [[0, 1], [2, 3], [4, 5]],
                dims=("x", "y"),
                coords={
                    "x": ("x", [10, 20, 30], {"units": unit_registry.Unit("dm")}),
                    "y": ("y", [60, 120], {"units": unit_registry.Unit("s")}),
                },
            ),
            xr.Dataset(
                {
                    "x": ("x", [1, 3, 5], {"units": unit_registry.Unit("m")}),
                    "y": ("y", [0, 2], {"units": unit_registry.Unit("min")}),
                }
            ),
            xr.DataArray(
                [[np.nan, 1], [np.nan, 5], [np.nan, np.nan]],
                dims=("x", "y"),
                coords={
                    "x": ("x", [1, 3, 5], {"units": unit_registry.Unit("m")}),
                    "y": ("y", [0, 2], {"units": unit_registry.Unit("min")}),
                },
            ),
            None,
            id="DataArray-compatible units",
        ),
        pytest.param(
            xr.DataArray(
                [[0, 1], [2, 3], [4, 5]],
                dims=("x", "y"),
                coords={
                    "x": ("x", [10, 20, 30], {"units": unit_registry.Unit("dm")}),
                    "y": ("y", [60, 120], {"units": unit_registry.Unit("s")}),
                },
            ),
            xr.Dataset(
                {
                    "x": ("x", [10, 30], {"units": unit_registry.Unit("s")}),
                    "y": ("y", [60], {"units": unit_registry.Unit("m")}),
                }
            ),
            None,
            ValueError,
            id="DataArray-incompatible units",
        ),
        pytest.param(
            xr.DataArray(
                Quantity([[0, 1], [2, 3], [4, 5]], "kg"),
                dims=("x", "y"),
                coords={
                    "x": [10, 20, 30],
                    "y": [60, 120],
                },
            ),
            xr.Dataset(
                {
                    "x": [15, 25],
                    "y": [75, 105],
                }
            ),
            xr.DataArray(
                Quantity([[1.25, 1.75], [3.25, 3.75]], "kg"),
                dims=("x", "y"),
                coords={
                    "x": [15, 25],
                    "y": [75, 105],
                },
            ),
            None,
            id="DataArray-data units",
        ),
    ),
)
def test_interp_like(obj, other, expected, error):
    if error is not None:
        with pytest.raises(error):
            obj.pint.interp_like(other)
    else:
        actual = obj.pint.interp_like(other)
        assert_units_equal(actual, expected)
        assert_identical(actual, expected)


@requires_bottleneck
@pytest.mark.parametrize(
    ["obj", "expected"],
    (
        pytest.param(
            xr.Dataset(
                {"a": ("x", [nan, 0, nan, 1, nan, nan, 2, nan])},
                coords={"u": ("x", [nan, 0, nan, 1, nan, nan, 2, nan])},
            ),
            xr.Dataset(
                {"a": ("x", [nan, 0, 0, 1, 1, 1, 2, 2])},
                coords={"u": ("x", [nan, 0, nan, 1, nan, nan, 2, nan])},
            ),
            id="Dataset-no units",
        ),
        pytest.param(
            xr.Dataset(
                {
                    "a": (
                        "x",
                        Quantity([nan, 0, nan, 1, nan, nan, 2, nan], "m"),
                    )
                },
                coords={
                    "u": (
                        "x",
                        Quantity([nan, 0, nan, 1, nan, nan, 2, nan], "m"),
                    )
                },
            ),
            xr.Dataset(
                {"a": ("x", Quantity([nan, 0, 0, 1, 1, 1, 2, 2], "m"))},
                coords={
                    "u": (
                        "x",
                        Quantity([nan, 0, nan, 1, nan, nan, 2, nan], "m"),
                    )
                },
            ),
            id="Dataset-units",
        ),
        pytest.param(
            xr.DataArray(
                [nan, 0, nan, 1, nan, nan, 2, nan],
                coords={
                    "u": (
                        "x",
                        [nan, 0, nan, 1, nan, nan, 2, nan],
                    )
                },
                dims="x",
            ),
            xr.DataArray(
                [nan, 0, 0, 1, 1, 1, 2, 2],
                coords={
                    "u": (
                        "x",
                        [nan, 0, nan, 1, nan, nan, 2, nan],
                    )
                },
                dims="x",
            ),
            id="DataArray-no units",
        ),
        pytest.param(
            xr.DataArray(
                Quantity([nan, 0, nan, 1, nan, nan, 2, nan], "m"),
                coords={
                    "u": (
                        "x",
                        Quantity([nan, 0, nan, 1, nan, nan, 2, nan], "m"),
                    )
                },
                dims="x",
            ),
            xr.DataArray(
                Quantity([nan, 0, 0, 1, 1, 1, 2, 2], "m"),
                coords={
                    "u": (
                        "x",
                        Quantity([nan, 0, nan, 1, nan, nan, 2, nan], "m"),
                    )
                },
                dims="x",
            ),
            id="DataArray-units",
        ),
    ),
)
def test_ffill(obj, expected):
    actual = obj.pint.ffill(dim="x")
    assert_identical(actual, expected)
    assert_units_equal(actual, expected)


@requires_bottleneck
@pytest.mark.parametrize(
    ["obj", "expected"],
    (
        pytest.param(
            xr.Dataset(
                {"a": ("x", [nan, 0, nan, 1, nan, nan, 2, nan])},
                coords={"u": ("x", [nan, 0, nan, 1, nan, nan, 2, nan])},
            ),
            xr.Dataset(
                {"a": ("x", [0, 0, 1, 1, 2, 2, 2, nan])},
                coords={"u": ("x", [nan, 0, nan, 1, nan, nan, 2, nan])},
            ),
            id="Dataset-no units",
        ),
        pytest.param(
            xr.Dataset(
                {
                    "a": (
                        "x",
                        Quantity([nan, 0, nan, 1, nan, nan, 2, nan], "m"),
                    )
                },
                coords={
                    "u": (
                        "x",
                        Quantity([nan, 0, nan, 1, nan, nan, 2, nan], "m"),
                    )
                },
            ),
            xr.Dataset(
                {"a": ("x", Quantity([0, 0, 1, 1, 2, 2, 2, nan], "m"))},
                coords={
                    "u": (
                        "x",
                        Quantity([nan, 0, nan, 1, nan, nan, 2, nan], "m"),
                    )
                },
            ),
            id="Dataset-units",
        ),
        pytest.param(
            xr.DataArray(
                [nan, 0, nan, 1, nan, nan, 2, nan],
                coords={
                    "u": (
                        "x",
                        [nan, 0, nan, 1, nan, nan, 2, nan],
                    )
                },
                dims="x",
            ),
            xr.DataArray(
                [0, 0, 1, 1, 2, 2, 2, nan],
                coords={
                    "u": (
                        "x",
                        [nan, 0, nan, 1, nan, nan, 2, nan],
                    )
                },
                dims="x",
            ),
            id="DataArray-no units",
        ),
        pytest.param(
            xr.DataArray(
                Quantity([nan, 0, nan, 1, nan, nan, 2, nan], "m"),
                coords={
                    "u": (
                        "x",
                        Quantity([nan, 0, nan, 1, nan, nan, 2, nan], "m"),
                    )
                },
                dims="x",
            ),
            xr.DataArray(
                Quantity([0, 0, 1, 1, 2, 2, 2, nan], "m"),
                coords={
                    "u": (
                        "x",
                        Quantity([nan, 0, nan, 1, nan, nan, 2, nan], "m"),
                    )
                },
                dims="x",
            ),
            id="DataArray-units",
        ),
    ),
)
def test_bfill(obj, expected):
    actual = obj.pint.bfill(dim="x")
    assert_identical(actual, expected)
    assert_units_equal(actual, expected)


@pytest.mark.parametrize(
    ["obj", "expected"],
    (
        pytest.param(
            xr.Dataset(
                {"a": ("x", [nan, 0, nan, 1, nan, nan, nan, 2, nan])},
                coords={"u": ("x", [nan, 0, nan, 1, nan, nan, nan, 2, nan])},
            ),
            xr.Dataset(
                {"a": ("x", [nan, 0, 0.5, 1, 1.25, 1.5, 1.75, 2, nan])},
                coords={"u": ("x", [nan, 0, nan, 1, nan, nan, nan, 2, nan])},
            ),
            id="Dataset-no units",
        ),
        pytest.param(
            xr.Dataset(
                {
                    "a": (
                        "x",
                        Quantity([nan, 0, nan, 1, nan, nan, nan, 2, nan], "m"),
                    )
                },
                coords={
                    "u": (
                        "x",
                        Quantity([nan, 0, nan, 1, nan, nan, nan, 2, nan], "m"),
                    )
                },
            ),
            xr.Dataset(
                {"a": ("x", Quantity([nan, 0, 0.5, 1, 1.25, 1.5, 1.75, 2, nan], "m"))},
                coords={
                    "u": (
                        "x",
                        Quantity([nan, 0, nan, 1, nan, nan, nan, 2, nan], "m"),
                    )
                },
            ),
            id="Dataset-units",
        ),
        pytest.param(
            xr.DataArray(
                [nan, 0, nan, 1, nan, nan, nan, 2, nan],
                coords={
                    "u": (
                        "x",
                        Quantity([nan, 0, nan, 1, nan, nan, nan, 2, nan], "m"),
                    )
                },
                dims="x",
            ),
            xr.DataArray(
                [nan, 0, 0.5, 1, 1.25, 1.5, 1.75, 2, nan],
                coords={
                    "u": (
                        "x",
                        Quantity([nan, 0, nan, 1, nan, nan, nan, 2, nan], "m"),
                    )
                },
                dims="x",
            ),
            id="DataArray-units",
        ),
        pytest.param(
            xr.DataArray(
                Quantity([nan, 0, nan, 1, nan, nan, nan, 2, nan], "m"),
                coords={
                    "u": (
                        "x",
                        Quantity([nan, 0, nan, 1, nan, nan, nan, 2, nan], "m"),
                    )
                },
                dims="x",
            ),
            xr.DataArray(
                Quantity([nan, 0, 0.5, 1, 1.25, 1.5, 1.75, 2, nan], "m"),
                coords={
                    "u": (
                        "x",
                        Quantity([nan, 0, nan, 1, nan, nan, nan, 2, nan], "m"),
                    )
                },
                dims="x",
            ),
            id="DataArray-units",
        ),
    ),
)
def test_interpolate_na(obj, expected):
    actual = obj.pint.interpolate_na(dim="x")
    assert_identical(actual, expected)
    assert_units_equal(actual, expected)



import re
from contextlib import contextmanager

import numpy as np
import pytest
from pint import Quantity
from xarray import DataArray, Variable
from xarray.testing import assert_equal, assert_identical  # noqa: F401

from ..conversion import (
    array_strip_units,
    extract_indexer_units,
    strip_units,
    strip_units_variable,
)
from ..testing import assert_units_equal  # noqa: F401


def importorskip(name):
    try:
        __import__(name)
        has_name = True
    except ImportError:
        has_name = False

    return has_name, pytest.mark.skipif(not has_name, reason=f"{name} is not available")


has_dask_array, requires_dask_array = importorskip("dask.array")
has_scipy, requires_scipy = importorskip("scipy")
has_bottleneck, requires_bottleneck = importorskip("bottleneck")


@contextmanager
def raises_regex(error, pattern):
    __tracebackhide__ = True
    with pytest.raises(error) as excinfo:
        yield
    message = str(excinfo.value)
    if not re.search(pattern, message):
        raise AssertionError(
            f"exception {excinfo.value!r} did not match pattern {pattern!r}"
        )


def assert_array_units_equal(a, b):
    __tracebackhide__ = True

    units_a = getattr(a, "units", None)
    units_b = getattr(b, "units", None)

    assert units_a == units_b


def assert_array_equal(a, b):
    __tracebackhide__ = True

    a_ = getattr(a, "magnitude", a)
    b_ = getattr(b, "magnitude", b)

    np.testing.assert_array_equal(a_, b_)


def assert_slice_equal(a, b):
    attrs = ("start", "stop", "step")
    values_a = tuple(getattr(a, name) for name in attrs)
    values_b = tuple(getattr(b, name) for name in attrs)
    stripped_a = tuple(array_strip_units(v) for v in values_a)
    stripped_b = tuple(array_strip_units(v) for v in values_b)

    assert (
        stripped_a == stripped_b
    ), f"different values: {stripped_a!r} ←→ {stripped_b!r}"


def assert_indexer_equal(a, b):
    __tracebackhide__ = True

    assert type(a) == type(b)
    if isinstance(a, slice):
        assert_slice_equal(a, b)
    elif isinstance(a, DataArray):
        stripped_a = strip_units(a)
        stripped_b = strip_units(b)

        assert_equal(stripped_a, stripped_b)
    elif isinstance(a, Variable):
        stripped_a = strip_units_variable(a)
        stripped_b = strip_units_variable(b)

        assert_equal(stripped_a, stripped_b)
    elif isinstance(a, (Quantity, np.ndarray)):
        assert_array_equal(a, b)
    else:
        a_ = array_strip_units(a)
        b_ = array_strip_units(b)
        assert a_ == b_, f"different values: {a_!r} ←→ {b_!r}"


def assert_indexer_units_equal(a, b):
    __tracebackhide__ = True

    units_a = extract_indexer_units(a)
    units_b = extract_indexer_units(b)

    assert units_a == units_b, f"different units: {units_a!r} ←→ {units_b!r}"



# type: ignore
import argparse
import functools
import json
import pathlib
import textwrap
from dataclasses import dataclass

from pytest import CollectReport, TestReport


@dataclass
class SessionStart:
    pytest_version: str
    outcome: str = "status"

    @classmethod
    def _from_json(cls, json):
        json_ = json.copy()
        json_.pop("$report_type")
        return cls(**json_)


@dataclass
class SessionFinish:
    exitstatus: str
    outcome: str = "status"

    @classmethod
    def _from_json(cls, json):
        json_ = json.copy()
        json_.pop("$report_type")
        return cls(**json_)


def parse_record(record):
    report_types = {
        "TestReport": TestReport,
        "CollectReport": CollectReport,
        "SessionStart": SessionStart,
        "SessionFinish": SessionFinish,
    }
    cls = report_types.get(record["$report_type"])
    if cls is None:
        raise ValueError(f"unknown report type: {record['$report_type']}")

    return cls._from_json(record)


@functools.singledispatch
def format_summary(report):
    return f"{report.nodeid}: {report}"


@format_summary.register
def _(report: TestReport):
    message = report.longrepr.chain[0][1].message
    return f"{report.nodeid}: {message}"


@format_summary.register
def _(report: CollectReport):
    message = report.longrepr.split("\n")[-1].removeprefix("E").lstrip()
    return f"{report.nodeid}: {message}"


def format_report(reports, py_version):
    newline = "\n"
    summaries = newline.join(format_summary(r) for r in reports)
    message = textwrap.dedent(
        """\
        <details><summary>Python {py_version} Test Summary</summary>

        ```
        {summaries}
        ```

        </details>
        """
    ).format(summaries=summaries, py_version=py_version)
    return message


if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument("filepath", type=pathlib.Path)
    args = parser.parse_args()

    py_version = args.filepath.stem.split("-")[1]

    print("Parsing logs ...")

    lines = args.filepath.read_text().splitlines()
    reports = [parse_record(json.loads(line)) for line in lines]

    failed = [report for report in reports if report.outcome == "failed"]

    message = format_report(failed, py_version=py_version)

    output_file = pathlib.Path("pytest-logs.txt")
    print(f"Writing output file to: {output_file.absolute()}")
    output_file.write_text(message)
